<?php

namespace Drupal\ekan_datastore;

/**
 * Class for handling setting/getting key values with locking.
 */
class LockableDrupalVariables {

  /**
   * The name of the key value store where the locks are kept.
   *
   * @var string
   */
  private string $binName;

  /**
   * The locking class.
   *
   * @var \Drupal\ekan_datastore\Locker
   */
  private Locker $locker;

  /**
   * LockableDrupalVariables constructor.
   */
  public function __construct($bin_name) {
    $this->binName = $bin_name;
    $this->locker = new Locker();
  }

  /**
   * Get all the variables in this bin.
   *
   * This operation will lock other bin operataions until the
   * return bin method is called.
   *
   * @return array
   *   All of the variables in this bin.
   */
  public function borrowBin() {
    $this->locker->getLock();
    $bin = \Drupal::keyValue($this->binName)->getAll();
    return $bin;
  }

  /**
   * Return bin.
   *
   * Sets the whole bin and all of its variables.
   * Releases the lock set if borrowBin() was called.
   *
   * @param array $bin
   *   The full bin.
   */
  public function returnBin(array $bin) {
    \Drupal::keyValue($this->binName)->setMultiple($bin);
    $this->locker->releaseLock();
  }

  /**
   * Set a variable.
   *
   * @param string $id
   *   The variable's id.
   * @param mixed $data
   *   The variable's value.
   */
  public function set($id, $data) {
    $this->locker->getLock();
    \Drupal::keyValue($this->binName)->set($id, $data);
    $this->locker->releaseLock();
  }

  /**
   * Get the variable with the given id.
   *
   * @param string $id
   *   The variable's id.
   *
   * @return mixed
   *   The variable's value.
   */
  public function get($id) {
    $this->locker->getLock();
    $var = \Drupal::keyValue($this->binName)->get($id);
    $this->locker->releaseLock();
    return $var ?? NULL;
  }

  /**
   * Delete the variable with the given id.
   *
   * @param string $id
   *   The id of the variable.
   */
  public function delete($id) {
    $this->locker->getLock();
    \Drupal::keyValue($this->binName)->delete($id);
    $this->locker->releaseLock();
  }

}
